/*
 *
 *	Utility for editting quotas
 *
 */

#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <stdlib.h>
#include <signal.h>
#include <errno.h>
#include <paths.h>

#include <linux/quota.h>
#include <sys/wait.h>
#include <sys/stat.h>

#include "common.h"
#include "quota_io.h"
#include "quotasys.h"

#define VI_PATH "vi"
#define MAXLINELEN 256

#define FL_COPY 1
#define FL_GRACE 2

struct dev_mnt {
  char dev[MAXDEVLEN];
  char mnt[MAXFILENAMELEN];
  struct dev_mnt *next;
};

int type, flags;
int names;
qid_t protoid;

void usage(char *utilname)
{
	printf("Utility for editting user/group quotas.\nUsage:\n%s [-u | -g] [-p modelname] name...\n", utilname);
	printf("%s -t [-u | -g]\n", utilname);
	printf("Bugs to %s\n", MY_EMAIL);
	exit(1);
}

void parse_options(int argcnt, char **argstr)
{
	int ret;
	char *slash = strrchr(argstr[0], '/');

	if (!slash)
		slash = argstr[0];
	else
		slash++;

	type = QS_USRQUOTA;
	while ((ret = getopt(argcnt, argstr, "Vughtp:")) != EOF) {
		switch (ret) {
			case '?':
			case 'h':
				usage(slash);
			case 'V':
				version();
				exit(0);
			case 'g':
				type = QS_GRPQUOTA;
				break;
			case 'u':
				type = QS_USRQUOTA;
				break;
			case 't':
				flags |= FL_GRACE;
				break;
			case 'p':
				flags |= FL_COPY;
				if (type == QS_USRQUOTA)
					protoid = name2uid(optarg);
				else
					protoid = group2gid(optarg);
				break;
		}
	}
	if (flags & FL_GRACE && flags & FL_COPY) {
		fputs("Illegal combination of options.\n", stderr);
		usage(slash);
	}
	if ((flags & FL_GRACE && optind != argcnt) || (!(flags & FL_GRACE) && optind >= argcnt)) {
		fputs("Bad number of arguments.\n", stderr);
		usage(slash);
	}
	names = optind;
}

void print_graces(FILE *tf)
{
	char bgbuf[MAXTIMELEN], igbuf[MAXTIMELEN];

	time2str(info.dqi_bgrace, bgbuf, 0);
	time2str(info.dqi_igrace, igbuf, 0);
#ifdef ALT_FORMAT
	fprintf(tf, "%20s %20s\n", bgbuf, igbuf);
#else
	fprintf(tf, "Block grace: %s Inode grace: %s\n", bgbuf, igbuf);
#endif
}

void print_limits(FILE *tf, qid_t id)
{
	struct dquot dquot;

	read_dquot(id, &dquot);
#ifdef ALT_FORMAT
	fprintf(tf, "%10lu %10lu %10lu %9lu %9lu %9lu\n", (ulong)(dquot.dq_dqb.dqb_curspace>>10),
	  (ulong)dquot.dq_dqb.dqb_bsoftlimit, (ulong)dquot.dq_dqb.dqb_bhardlimit, (ulong)dquot.dq_dqb.dqb_curinodes,
	  (ulong)dquot.dq_dqb.dqb_isoftlimit, (ulong)dquot.dq_dqb.dqb_ihardlimit);
#else
	fprintf(tf, "Used %luKB, limits: soft=%lu hard=%lu\n", (ulong)(dquot.dq_dqb.dqb_curspace>>10),
	  (ulong)dquot.dq_dqb.dqb_bsoftlimit, (ulong)dquot.dq_dqb.dqb_bhardlimit);
	fprintf(tf, "Used %lu inodes, limits: soft=%lu hard=%lu\n", (ulong)dquot.dq_dqb.dqb_curinodes,
	  (ulong)dquot.dq_dqb.dqb_isoftlimit, (ulong)dquot.dq_dqb.dqb_ihardlimit);
#endif
}

long parse_date(char *buf)
{
	char *errch;
	long tn;

	tn = strtol(buf, &errch, 10);
	if (*errch == ' ' || *errch == 's' || *errch == 'S')
		return tn;
	if (!strcasecmp(errch, "minutes") || !strcasecmp(errch, "min"))
		return tn*60;
	if (!strcasecmp(errch, "hours") || !strcasecmp(errch, "hour"))
		return tn*3600;
	if (!strcasecmp(errch, "days") || !strcasecmp(errch, "day"))
		return tn*86400;
	fprintf(stderr, "Can't parse grace time: %s\n", buf);
	return -1;
}

int scan_graces(FILE *tf)
{
	char linebuf[MAXLINELEN], btimes[MAXLINELEN], itimes[MAXLINELEN];
	long timeret;

	if (!fgets(linebuf, MAXLINELEN, tf)) {
		fprintf(stderr, "Unexpected end of file.\n");
		return 1;
	}
#ifdef ALT_FORMAT
	sscanf(linebuf, "%s %s", btimes, itimes);
#else
	sscanf(linebuf, "Block grace: %s Inode grace: %s", btimes, itimes);
#endif
	timeret = parse_date(btimes);
	if (timeret >= 0)
		info.dqi_bgrace = timeret;
	else
		return 1;
	timeret = parse_date(itimes);
	if (timeret >= 0)
		info.dqi_igrace = timeret;
	else
		return 1;
	mark_quotafile_info_dirty();
	return 0;
}

int scan_limits(FILE *tf, qid_t id)
{
	char linebuf[MAXLINELEN];
	struct dquot dquot;
	ulong busage, bsoftlim, bhardlim, iusage, isoftlim, ihardlim;

	read_dquot(id, &dquot);
#ifdef ALT_FORMAT
	if (!fgets(linebuf, MAXLINELEN, tf)) {
		fprintf(stderr, "Unexpected end of file.\n");
		return 1;
	}
	if (sscanf(linebuf, "%lu %lu %lu %lu %lu %lu", &busage, &bsoftlim, &bhardlim, &iusage, &isoftlim, &ihardlim) != 6) {
		fprintf(stderr, "Can't parse limits: %s\n", linebuf);
		return 1;
	}
	dquot.dq_dqb.dqb_bsoftlimit = bsoftlim;
	dquot.dq_dqb.dqb_bhardlimit = bhardlim;
	dquot.dq_dqb.dqb_isoftlimit = isoftlim;
	dquot.dq_dqb.dqb_ihardlimit = ihardlim;
#else
	if (!fgets(linebuf, MAXLINELEN, tf)) {
bad_eof:
		fprintf(stderr, "Unexpected end of file.\n");
		return 1;
	}
	if (sscanf(linebuf, "Used %luKB, limits: soft=%lu hard=%lu", &busage, &bsoftlim, &bhardlim) != 3) {
bad_limits:
		fprintf(stderr, "Can't parse limits: %s\n", linebuf);
		return 1;
	}
	dquot.dq_dqb.dqb_bsoftlimit = bsoftlim;
	dquot.dq_dqb.dqb_bhardlimit = bhardlim;
	if (!fgets(linebuf, MAXLINELEN, tf))
		goto bad_eof;
	if (sscanf(linebuf, "Used %lu inodes, limits: soft=%lu hard=%lu", &iusage, &isoftlim, &ihardlim) != 3)
		goto bad_limits;
	dquot.dq_dqb.dqb_isoftlimit = isoftlim;
	dquot.dq_dqb.dqb_ihardlimit = ihardlim;
#endif
	update_times(&dquot.dq_dqb);
	commit_dquot(&dquot, 0);
	return 0;
}

int scan_mountpoint(FILE *tf, char *devname)
{
#ifdef ALT_FORMAT
	int ret;

	if (!(ret = fscanf(tf, "%24s", devname))) {
		fprintf(stderr, "Illegal format of edited file. Can't parse device name.\n");
		return 1;
	}
	if (ret == -1)
		return -1;
	return 0;
#else
	char linebuf[MAXLINELEN];
	char *ch1, *ch2;

	if (!fgets(linebuf, MAXLINELEN, tf))
		return -1;
	if (!(ch1 = strchr(linebuf, ' ')))
		goto out;
	if (!(ch2 = strchr(++ch1, ' ')))
		goto out;
	*ch2 = 0;
	sstrncpy(devname, ch1, MAXDEVLEN);
	return 0;
out:
	fprintf(stderr, "Illegal format of edited file. Can't parse: %s\n", linebuf);
	return 1;
#endif
}

void copy_quota(int type, qid_t editid)
{
	char devname[MAXDEVLEN];
	char mntpoint[MAXFILENAMELEN];

	init_mount(NULL, 0);
	while (next_mount(type, devname, mntpoint, NULL)) {
		struct dquot dquot, adquot;

		init_io(mntpoint, devname, type);
		read_dquot(protoid, &dquot);
		read_dquot(editid, &adquot);
		adquot.dq_dqb.dqb_bsoftlimit = dquot.dq_dqb.dqb_bsoftlimit;
		adquot.dq_dqb.dqb_bhardlimit = dquot.dq_dqb.dqb_bhardlimit;
		adquot.dq_dqb.dqb_isoftlimit = dquot.dq_dqb.dqb_isoftlimit;
		adquot.dq_dqb.dqb_ihardlimit = dquot.dq_dqb.dqb_ihardlimit;
		update_times(&adquot.dq_dqb);
		commit_dquot(&adquot, 0);
		free_io();
	}
	free_mount();
}

void edit(char *filename)
{
	int pid, stat;

	if ((pid = fork()) < 0)
		die(11, "Canot fork to run editor: %s\n", strerror(errno));
	if (!pid) {
		char *editor;

		if (!(editor = getenv("VISUAL")) && !(editor = getenv("EDITOR")))
			editor = VI_PATH;
		if (execlp(editor, editor, filename, NULL) < 0)
			die(11, "Can't run editor %s: %s\n", editor, strerror(errno));
	}
	waitpid(pid, &stat, 0);
}

int edit_again(void)
{
	char buf[16];

	printf("Do you want to edit file again (y/n)? ");
	fflush(stdout);
	fgets(buf, sizeof(buf), stdin);
	return !strcasecmp(buf, "y\n");
}

void process_name(int type, char *name)
{
	qid_t id = 0;
	int err;

	if (!(flags & FL_GRACE))
		if (type == QS_USRQUOTA)
			id = name2uid(name);
		else
			id = group2gid(name);

	if (flags & FL_COPY)
		copy_quota(type, id);
	else {
		FILE *tmp;
		int tmpfd;
		char devname[MAXDEVLEN];
		char mntpoint[MAXFILENAMELEN];
		char tmpname[MAXFILENAMELEN];
		char linebuf[MAXLINELEN];
		struct dev_mnt *fmd = NULL, **lmd = &fmd, *nmd;

		umask(077);
		sstrncpy(tmpname, _PATH_TMP, MAXFILENAMELEN);
		sstrncat(tmpname, "edqXXXXXX", MAXFILENAMELEN);
		if ((tmpfd = mkstemp(tmpname)) < 0)
			die(10, "Can't create temporary file: %s\n", strerror(errno));
		if (!(tmp = fdopen(tmpfd, "r+")))
			die(10, "Can't reopen temporary file: %s\n", strerror(errno));
#ifdef ALT_FORMAT
		if (flags & FL_GRACE)
			fprintf(tmp, "Grace period before enforcing soft limit for %ss\n\
  Filesystem                Block grace period   Inode grace period\n", type2name(type));
		else
			fprintf(tmp, "Quotas for %s %s:\n\
  Filesystem                  Blocks       Soft       Hard    Inodes      Soft      Hard\n", type2name(type), name);
#else
		if (flags & FL_GRACE)
			fprintf(tmp, "Edit grace times for %s quota:\n", type2name(type));
		else
			fprintf(tmp, "Edit block and inode quota for %s %s:\n", type2name(type), name);
#endif
		init_mount(NULL, 0);
		while (next_mount(type, devname, mntpoint, NULL)) {
			init_io(mntpoint, devname, type);
			/* Create structure with device<->mountpoint link */
			*lmd = smalloc(sizeof(struct dev_mnt));
			strcpy((*lmd)->dev, devname);
			strcpy((*lmd)->mnt, mntpoint);
			(*lmd)->next = NULL;
			lmd = &((*lmd)->next);
#ifdef ALT_FORMAT
			fprintf(tmp, "%-26s", devname);
#else
			fprintf(tmp, "Device %s (%s):\n", devname, mntpoint);
#endif
			if (flags & FL_GRACE)
				print_graces(tmp);
			else
				print_limits(tmp, id);
			free_io();
		}
		free_mount();
		fclose(tmp);
		do {
			err = 0;
			edit(tmpname);
			if (!(tmp = fopen(tmpname, "r")))
				die(10, "Can't reopen temporary file %s: %s\n", tmpname, strerror(errno));
			fgets(linebuf, MAXLINELEN, tmp);	/* Skip first line */
#ifdef ALT_FORMAT
			fgets(linebuf, MAXLINELEN, tmp);	/* Discard one more line */
#endif
			while (!err) {
				if ((err = scan_mountpoint(tmp, devname)) != 0) /* EOF / error? */
					break;
				if (strcmp(devname, fmd->dev)) {
					fprintf(stderr, "Device names doesn't match (%s != %s).\n", devname, fmd->dev);
					err = -1;
					break;
				}
				strcpy(mntpoint, fmd->mnt);
				nmd = fmd->next;
				free(fmd);
				fmd = nmd;
				init_io(mntpoint, devname, type);
				if (flags & FL_GRACE)
					err = scan_graces(tmp);
				else
					err = scan_limits(tmp, id);
				free_io();
			}
			fclose(tmp);
		} while (err > 0 && edit_again());
		unlink(tmpname);
	}
}

int main(int argcnt, char **argstr)
{
	parse_options(argcnt, argstr);
	if (flags & FL_GRACE)
		process_name(type, NULL);
	else
		for (;names < argcnt; names++)
			process_name(type, argstr[names]);
	return 0;
}
