/*
 *
 *	Interactions of quota with system - filenames, fstab and so on...
 *
 */

#include <stdio.h>
#include <mntent.h>
#include <errno.h>
#include <string.h>
#include <pwd.h>
#include <grp.h>
#include <stdlib.h>
#include <time.h>

#include <linux/quota.h>

#include "common.h"
#include "quotasys.h"

static char names[MAXQUOTAS + 2][20] = INITQFNAMES;

/*
 *	Convert our representation of quota types to kernel one
 */
int user2kernquota(int x)
{
	if (x & QS_USRQUOTA)
		return USRQUOTA;
	else if (x & QS_GRPQUOTA)
		return GRPQUOTA;
	else
		return -1;
}

/*
 *	Convert kernel representation of quota types to our
 */
int kern2userquota(int x)
{
	return 1 << x;
}

/*
 *	Convert type of quota to written representation
 */
char *type2name(int type)
{
	return names[user2kernquota(type)];
}

/*
 *	Get name of quota file
 */
void get_qfname(char *mntpoint, int type, char *buf)
{
	int kernt = user2kernquota(type);

	sstrncpy(buf, mntpoint, MAXFILENAMELEN);
	if (buf[strlen(buf) - 1] != '/')
		sstrncat(buf, "/", MAXFILENAMELEN);
	sstrncat(buf, QUOTAFILENAME, MAXFILENAMELEN);
	sstrncat(buf, ".", MAXFILENAMELEN);
	sstrncat(buf, names[kernt], MAXFILENAMELEN);
}

FILE *mtab;
char **mntpoints;
int mntcnt;
char mnt_done[MAXMNTPOINTS];

/*
 *	Prepare for searching in /etc/mtab
 */
void init_mount(char **umntpoints, int umntcnt)
{
	if (!(mtab = setmntent("/etc/mtab", "r"))) {
		perror("Can't open /etc/mtab");
		exit(2);
	}
	mntpoints = umntpoints;
	if (umntcnt > MAXMNTPOINTS)
		die(1, "Too many mount points. Maximum is %d.\n", MAXMNTPOINTS);
	mntcnt = umntcnt;
	memset(mnt_done, 0, MAXMNTPOINTS);
}

static int hasquotaopt(char *opts, int type)
{
	char *qopt;
	char *actchr = opts, *nextchr;
	int len;

	if (type & QS_USRQUOTA)
		qopt = "usrquota";
	else if (type & QS_GRPQUOTA)
		qopt = "grpquota";
	else
		qopt = "unknown";
	len = strlen(qopt);
	for (nextchr = strchr(actchr, ','); nextchr; (actchr = nextchr + 1), nextchr = strchr(nextchr+1, ',')) {
		if (nextchr - actchr != len)
			continue;
		if (!strncasecmp(actchr, qopt, len))
			return 1;
	}
	return !strcasecmp(actchr, qopt);
}

/*
 *	Return next mountpoint which concerns us
 */
int next_mount(int type, char *devbuf, char *mntbuf, char *fstypebuf)
{
	struct mntent *entry;
	int ret, i;

	while (1) {
		if (!(entry = getmntent(mtab)))
			return 0;
		for (i = 0; i < mntcnt && pathcmp(mntpoints[i], entry->mnt_dir); i++);
		if (mntcnt && i == mntcnt)
			continue;
		ret = hasquotaopt(entry->mnt_opts, QS_USRQUOTA) * QS_USRQUOTA;
		ret |= hasquotaopt(entry->mnt_opts, QS_GRPQUOTA) * QS_GRPQUOTA;
		if (!(ret & type))
			continue;
		sstrncpy(devbuf, entry->mnt_fsname, MAXDEVLEN);
		sstrncpy(mntbuf, entry->mnt_dir, MAXFILENAMELEN);
		if (fstypebuf)
			sstrncpy(fstypebuf, entry->mnt_type, MAXFSTYPELEN);
		if (mntcnt)
			mnt_done[i] = 1;
		return ret & type;
	}
}

/*
 *	Finish examining of /etc/mtab
 */
void free_mount(void)
{
	int i;

	endmntent(mtab);
	for (i = 0; i < mntcnt; i++)
		if (!mnt_done[i])
			fprintf(stderr, "Warning: Mountpoint %s wasn't found in /etc/mtab or isn't mounted with quotas.\n", mntpoints[i]);
}

/*
 *	Convert name to uid
 */

uid_t name2uid(char *name)
{
	struct passwd *entry;
	uid_t ret;
	char *errch;

	ret = strtol(name, &errch, 0);
	if (!*errch)	/* Is name number - we got directly uid? */
		return ret;
	if (!(entry = getpwnam(name))) {
		fprintf(stderr, "User %s doesn't exist.\n", name);
		exit(1);
	}
	return entry->pw_uid;
}

/*
 *	Convert group name to gid
 */
gid_t group2gid(char *name)
{
	struct group *entry;
	gid_t ret;
	char *errch;

	ret = strtol(name, &errch, 0);
	if (!*errch)	/* Is name number - we got directly gid? */
		return ret;
	if (!(entry = getgrnam(name))) {
		fprintf(stderr, "Group %s doesn't exist.\n", name);
		exit(1);
	}
	return entry->gr_gid;
}

/*
 *	Convert uid to name
 */
void uid2name(uid_t id, char *buf)
{
	struct passwd *entry;

	if (!(entry = getpwuid(id)))
		sprintf(buf, "#%u", (uint)id);
	else
		sstrncpy(buf, entry->pw_name, MAXNAMELEN);
}

/*
 *	Convert gid to name
 */
void gid2group(gid_t id, char *buf)
{
	struct group *entry;

	if (!(entry = getgrgid(id)))
		sprintf(buf, "#%u", (uint)id);
	else
		sstrncpy(buf, entry->gr_name, MAXNAMELEN);
}

/* Convert time difference of seconds and current time */
void difftime2str(time_t seconds, char *buf)
{
	time_t now;

	buf[0] = 0;
	if (!seconds)
		return;
	time(&now);
	if (seconds <= now) {
		strcpy(buf, "none");
		return;
	}
	time2str(seconds-now, buf, TF_ROUND);
}

/* Convert time to printable form */
void time2str(time_t seconds, char *buf, int flags)
{
	uint minutes, hours, days;

	minutes = (seconds + 30) / 60;  /* Rounding */
	hours = minutes / 60; minutes %= 60;
	days = hours / 24; hours %= 24;
	if (flags & TF_ROUND) {
		if (days)
			sprintf(buf, "%ddays", days);
		else
			sprintf(buf, "%02d:%02d", hours, minutes);
	}
	else {
		if (minutes || (!minutes && !hours && !days))
			sprintf(buf, "%uminutes", (uint)(seconds+30)/60);
		else if (hours)
			sprintf(buf, "%uhours", hours+days*24);
		else
			sprintf(buf, "%udays", days);
	}
}

